/*
 * $Id: starter_flash.cpp,v 1.7 2012/11/28 11:09:48 rp Exp $
 *
 * Starter Flash: Embed Flash in PDF
 *
 * Required software: PDFlib/PDFlib+PDI/PPS 9
 * Required data: Flash data files
 *
 * The Flash source files "mini.mxml" and "minivid.mxml" for the the Flash
 * video player "minivid.swf" and the Flash test application "mini.swf"
 * are included in the "data" directory.
 */

#include <iostream>

#include "pdflib.hpp"

using namespace std;
using namespace pdflib;

namespace {

/*
 * Utility function to create a button that calls an action on a rich-media
 * annotation.
 */
void
create_richmedia_action_button(PDFlib &p,
        const int font, const int fontsize,
        const wstring& caption,
        const wstring& tooltip,
        const wstring& target_name,
        const wstring& functionname,
        const int arg_handle,
        const double xpos, const double ypos,
        const double width, const double height)
{
    wostringstream optlist;

    optlist << L"target={" << target_name << L"} functionname=" << functionname;

    if (arg_handle != -1)
    {
        optlist << L" richmediaargs=" << arg_handle;
    }

    const int action = p.create_action(L"RichMediaExecute", optlist.str());

    optlist.str(L"");
    optlist << L"caption={" << caption << L"} "
                    << L"tooltip={" << tooltip << L"} "
                    << L"action={up=" << action << L"} "
                    << L"bordercolor={gray 0} "
                    << L"font=" << font << L" fontsize=" << fontsize;
    p.create_field(xpos, ypos, xpos + width, ypos + height,
            caption, L"pushbutton", optlist.str());
}

}

int
main(void)
{
    /* This is where the data files are. Adjust as necessary. */
    const wstring searchpath = L"../data";
    const wstring outfile = L"starter_flash.pdf";

    PDFlib p;

    try {
        const int pageheight = 600;
        const int pagewidth = 700;
        const int gap = pageheight / 20;

        const int fontsize = pageheight / 30;
        const int button_fontsize = (fontsize * 2) / 3;

        /* Layout for header */
        const int textbox_width = pagewidth - 2 * gap;
        const int textbox_height = fontsize;
        const int textbox_llx = gap;
        const int textbox_lly = pageheight - gap - fontsize;

        wostringstream optlist;
        int font, undirty_action;

	//  This means we must check return values of load_font() etc.
	p.set_option(L"errorpolicy=return");

	// Set the search path for fonts and PDF files
	optlist.str(L"");
	optlist << L"searchpath={{" << searchpath << L"}}";
	p.set_option(optlist.str());


        if (p.begin_document(outfile,
            L"viewerpreferences={fitwindow=true} compatibility=1.7ext3") == -1) {
            wcerr << L"Error: " << p.get_errmsg() << endl;
            return 2;
        }

        p.set_info(L"Creator", L"PDFlib starter sample");
        p.set_info(L"Title", L"starter_flash");

        font = p.load_font(L"Helvetica", L"winansi", L"");
        if (font == -1) {
            wcerr << L"Error: " << p.get_errmsg() << endl;
            return 2;
        }

        {
            /* Name of the flash video player application */
            const wstring video_player_file = L"minivid.swf";

            /* Name of the video to play */
            const wstring video_file = L"flashvid.flv";

            /* Layout for Flash area */
            const int flash_width = pagewidth - 2 * gap;
            const int flash_height = (flash_width * 10) / 18;
            const int flash_llx = textbox_llx;
            const int flash_lly = textbox_lly - gap - flash_height;

            /* Layout for pushbuttons */
            const int num_buttons = 5;
            const int button_gap = 20;
            const int button_width =
                (flash_width - (num_buttons - 1) * button_gap) / num_buttons;
            const int button_height = button_width / 2;
            const int button_llx = flash_llx;
            const int button_lly = flash_lly - gap - button_height;

#define NUM_CUEPOINTS 8
#define CUEPOINT_FIELDNAME L"cuepoint"
            const int textfield_gap = button_gap;
            const int textfield_llx = flash_llx;
            const int textfield_height = button_height / 2;
            const int textfield_lly = button_lly - gap - textfield_height;
            const int textfield_width =
                (flash_width - (NUM_CUEPOINTS - 1) * textfield_gap)
                                                        / NUM_CUEPOINTS;
            const int textfield_fontsize = (button_fontsize * 3) / 4;

            int cuepoint_actions[NUM_CUEPOINTS];
            const int cuepoint_timestamps[NUM_CUEPOINTS] =
            {
                /*
                 * The cue point timestamps do not have any functionality,
                 * but are displayed in Acrobat in the properties of the
                 * Flash video, therefore they should be accurate.
                 */
                660, 1133, 1700, 2266, 2933, 3500, 3933, 4566
            };

            int video_player, video, vol20pct_args, vol100pct_args, button_xpos,
                i, textfield_xpos, rm_action, js_action;

            /* Start a page */
            p.begin_page_ext(pagewidth, pageheight, L"");

            /* Header line */
            optlist.str(L"");
            optlist << L"position={center bottom} font=" << font
                    << L" fontsize=" << fontsize
                    << L" boxsize={" << textbox_width
                    << L" " << textbox_height << L"}";
#define PLAYER_TITLE \
    L"Video player with PDF button controls and cue point feedback"
            p.fit_textline(PLAYER_TITLE, textbox_llx, textbox_lly,
                            optlist.str());
            p.create_bookmark(PLAYER_TITLE, L"");

            /* Load the video display flash application */
            video_player = p.load_asset(L"Flash", video_player_file, L"");
            if (video_player == -1) {
                wcerr << L"Error: " << p.get_errmsg() << endl;
                return 2;
            }

            video = p.load_asset(L"Video", video_file,L"");
            if (video == -1) {
                wcerr << L"Error: " << p.get_errmsg() << endl;
                return 2;
            }

            /*
             * Create actions that are triggered by the cue points from the
             * Video. They change the visual appearance of the text fields at
             * the bottom of the page.
             */
            for (i = 0; i < NUM_CUEPOINTS; i += 1)
            {
                wostringstream javascript_buf;

                javascript_buf
                    << L"script {"
                        << L"var f = this.getField(\"" CUEPOINT_FIELDNAME << i << L"\"); "
                        << L"f.value = \"Cue point " << (i + 1) << L"\"; "
                        << L"f.fillColor = color.blue;"
                    << L"}";
                cuepoint_actions[i] = p.create_action(L"JavaScript",
                                            javascript_buf.str());
            }

            /*
             * Build the option list for creating the richmedia annotation. The
             * name for the video player and video assets can be freely choosen,
             * but must be consistent between the "assets" declarations and
             * the "instances" declarations.
             *
             * The cue points must be matchable by name to the actual cue points
             * in the Video. The "time" options do not have any effect, but it
             * is recommended to add them with proper values as Acrobat
             * displays them in the properties dialog of the video.
             */
#define VIDEO_PLAYER_ASSET_NAME L"minivid.swf"
#define VIDEO_ASSET_NAME L"flashvid.flv"
#define VIDEO_PLAYER_NAME L"Video Player"
            optlist.str(L"");
            optlist
                << L"name={" VIDEO_PLAYER_NAME L"} "
                << L"richmedia={"
                    << L"activate={condition=opened} "
                    << L"assets={"
                        << L"{asset=" << video_player << L" name={" VIDEO_PLAYER_ASSET_NAME L"}} "
                        << L"{asset=" << video << L" name={" VIDEO_ASSET_NAME L"}} "
                    << L"} "
                    << L"configuration={"
                    << L"type=Video name=DefaultConfiguration "
                        << L"instances={"
                            << L"{"
                                << L"asset=" VIDEO_PLAYER_ASSET_NAME L" "
                                << L"params={"
                                    << L"flashvars={"
                                        << L"source=" VIDEO_ASSET_NAME L"&"
                                         << L"title=PDFlib%20Video%20Player"
                                     << L"} "
                                     << L"cuepoints={";
                                     for (i = 0; i < NUM_CUEPOINTS; i += 1)
                                     {
                                         optlist
                                             << L"{"
                                                 << L"type=event "
                                                 << L"name={Cue-Point " << (i + 1) << L"} "
                                                 << L"time=" << cuepoint_timestamps[i] << L" "
                                                 << L"action={activate={" << cuepoint_actions[i] << L"}}"
                                             << L"} ";
                                     }
                                     optlist
                                     << L"} "
                                << L"}"
                            << L"}"
                        << L"} "
                    << L"} "
                << L"}";

            /* Create the rich-media annotation */
            p.create_annotation(flash_llx, flash_lly,
                    flash_llx + flash_width, flash_lly + flash_height,
                    L"RichMedia", optlist.str());

            /*
             * Create action buttons to drive the player
             */
            button_xpos = button_llx;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            L"Play", L"Play the video",
                            VIDEO_PLAYER_NAME, L"play", -1,
                            button_xpos, button_lly,
                            button_width, button_height);

            button_xpos += button_gap + button_width;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            L"Pause", L"Pause the video",
                            VIDEO_PLAYER_NAME, L"pause", -1,
                            button_xpos, button_lly,
                            button_width, button_height);

            /*
             * The rewind action is special because it not only executes a
             * RichMediaExecute action but also a JavaScript action to
             * reset the text fields for the cue point feedback.
             */
            button_xpos += button_gap + button_width;
            rm_action = p.create_action(L"RichMediaExecute",
                    L"target={" VIDEO_PLAYER_NAME L"} functionname=rewind");
            optlist.str(L"");
            optlist << L"script {"
                L"for (var i = 0; i < " << NUM_CUEPOINTS << L"; i += 1)"
                    L"{"
                        L"var f = this.getField(\"" CUEPOINT_FIELDNAME
                                                                L"\" + i); "
                        L"f.value = \"\"; "
                        L"f.fillColor = color.transparent;"
                    L"}"
                L"}";
            js_action = p.create_action(L"JavaScript", optlist.str());
            optlist.str(L"");
            optlist << L"caption={Rewind} "
                            << L"tooltip={Rewind the video} "
                            L"action={up={" << rm_action << L" " << js_action << L"}} "
                            L"bordercolor={gray 0} "
                            L"font=" << font
                            << L" fontsize=" << button_fontsize;
            p.create_field(button_xpos, button_lly,
                        button_xpos + button_width,
                        button_lly + button_height,
                        L"Rewind", L"pushbutton", optlist.str());

            /*
             * Create an argument list containing one argument to set the
             * volume for the video player to 20%. Note that Acrobat expects
             * floating point numbers as PDF type string.
             */
            vol20pct_args =
                    p.poca_new(L"containertype=array usage=richmediaargs");
            p.poca_insert(vol20pct_args,
                                    L"type=string index=0 value={0.2}");
            button_xpos += button_gap + button_width;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            L"Volume\n20%", L"Set volume to 20%",
                            VIDEO_PLAYER_NAME, L"volume", vol20pct_args,
                            button_xpos, button_lly,
                            button_width, button_height);

            /*
             * Create an argument list containing one argument to set the
             * volume for the video player to 100%.
             */
            vol100pct_args =
                p.poca_new(L"containertype=array usage=richmediaargs");
            p.poca_insert(vol100pct_args,
                                L"type=string index=0 value={1.0}");
            button_xpos += button_gap + button_width;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            L"Volume\n100%", L"Set volume to 100%",
                            VIDEO_PLAYER_NAME, L"volume", vol100pct_args,
                            button_xpos, button_lly,
                            button_width, button_height);

            /*
             * Create textfields for giving feedback about the cue points that
             * are reached.
             */
            for (i = 0, textfield_xpos = textfield_llx; i < NUM_CUEPOINTS;
                    i += 1, textfield_xpos += textfield_gap + textfield_width)
            {
                wostringstream fieldname;

                fieldname << CUEPOINT_FIELDNAME << i;

                optlist.str(L"");
                optlist << L"readonly=true font=" << font
                        << L" fontsize=" << textfield_fontsize
                            << L" fillcolor={gray 1} bordercolor={gray 0} "
                            << L"alignment=center";

                p.create_field(textfield_xpos, textfield_lly,
                        textfield_xpos + textfield_width,
                        textfield_lly + textfield_height,
                        fieldname.str(), L"textfield", optlist.str());
            }

            p.end_page_ext(L"");

            p.poca_delete(vol20pct_args, L"");
            p.poca_delete(vol100pct_args, L"");
        }

        {
            /* Name of the flash application */
            const wstring flash_app_file = L"mini.swf";

            /* Layout for Flash area */
            const int flash_width = 400;
            const int flash_height = 300;
            const int flash_llx = pagewidth / 2 - flash_width / 2;
            const int flash_lly = textbox_lly - gap - flash_height;

            /* Layout for pushbuttons */
            const int button_gap = 20;
            const int button_width = flash_width / 5;
            const int button_height = button_width * 2 / 3;
            const int button_llx =
                    pagewidth / 2 - button_gap / 2 - button_width;
            const int button_lly = flash_lly - gap - button_height;

            int flash_app, button_xpos, flash_args, flash_args2;

            /* Start a page */
            p.begin_page_ext(pagewidth, pageheight, L"");

            /* Header line */
            optlist.str(L"");
            optlist << L"position={center bottom} font=" << font
                    << L" fontsize=" << fontsize
                    << L" boxsize={" << textbox_width
                                    << L" " << textbox_height << L"}";
#define PLAYER_TITLE2 L"Passing parameters to functions in the Flash application"
            p.fit_textline(PLAYER_TITLE2, textbox_llx, textbox_lly,
                                optlist.str());
            p.create_bookmark(PLAYER_TITLE2, L"");

            /* Load the flash application */
            flash_app = p.load_asset(L"Flash", flash_app_file, L"");
            if (flash_app == -1) {
                wcerr << L"Error: " << p.get_errmsg() << endl;
                return 2;
            }

            /*
             * Build the option list for creating the richmedia annotation.
             */
#define FLASH_APP_NAME L"Flash Application"
            optlist.str(L"");
            optlist <<
                L"name={" FLASH_APP_NAME L"} "
                L"richmedia={"
                    L"assets={"
                        L"{asset=" << flash_app << L" name={" FLASH_APP_NAME L"}} "
                    L"} "
                    L"configuration={"
                        L"type=Flash "
                        L"instances={"
                            L"{"
                                L"asset={" FLASH_APP_NAME L"} "
                                L"params={"
                                    L"binding=foreground "
                                    L"flashvars={"
                                        L"title=Click below for test"
                                        L"&text="
                                    L"}"
                                L"}"
                            L"}"
                        L"} "
                    L"} "
                    L"activate={ "
                        L"condition=visible "
                    L"} "
                L"}";

            /* Create the rich-media annotation */
            p.create_annotation(flash_llx, flash_lly,
                    flash_llx + flash_width, flash_lly + flash_height,
                    L"RichMedia", optlist.str());

            /*
             * Create argument list for usage with varargs-style function.
             * This only works as expected with strings and booleans.
             */
            flash_args =
                    p.poca_new(L"containertype=array usage=richmediaargs");
            p.poca_insert(flash_args,
                    L"type=string index=0 value={Got varargs parameters}");
            p.poca_insert(flash_args,
                    L"type=string index=1 value={text string}");
            p.poca_insert(flash_args,
                    L"type=integer index=2 value={42}");
            p.poca_insert(flash_args,
                    L"type=float index=3 value={3.14159}");
            p.poca_insert(flash_args,
                    L"type=boolean index=4 value=true");
            p.poca_insert(flash_args,
                    L"type=string index=5 value={another string}");

            button_xpos = button_llx;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            L"Varargs",
                            L"Pass arguments to Flash application via "
                                    L"variable-length argument list",
                            FLASH_APP_NAME, L"myCallbackVarargs", flash_args,
                            button_xpos, button_lly,
                            button_width, button_height);

            /*
             * Create argument list for a function with types for the
             * parameters. Note that numbers must be passed as strings.
             */
            flash_args2 =
                    p.poca_new(L"containertype=array usage=richmediaargs");
            p.poca_insert(flash_args2,
                    L"type=string index=0 value={Got typed parameters}");
            p.poca_insert(flash_args2,
                    L"type=boolean index=1 value=false");
            p.poca_insert(flash_args2,
                    L"type=string index=2 value={3.14159}");
            p.poca_insert(flash_args2,
                    L"type=string index=3 value={42}");
            p.poca_insert(flash_args2,
                    L"type=string index=4 value={typed string argument}");

            button_xpos += button_gap + button_width;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            L"Typed\narguments",
                            L"Pass arguments to Flash application via typed "
                            L"argument list",
                            FLASH_APP_NAME, L"myCallbackTyped", flash_args2,
                            button_xpos, button_lly,
                            button_width, button_height);

            p.end_page_ext(L"");

            p.poca_delete(flash_args, L"");
            p.poca_delete(flash_args2, L"");
        }

        /*
         * Prevent the prompt for saving the document, that is caused by
         * modifying the text fields for the cue points.
         */
        undirty_action = p.create_action(L"JavaScript",
                            L"script={ this.dirty = false; }");
        optlist.str(L"");
        optlist << L"action={willclose=" << undirty_action << L"}";
        p.end_document(optlist.str());
    }
    catch (PDFlib::Exception &ex) {
        wcerr << L"PDFlib exception occurred in hello sample: " << endl
                      << L"[" << ex.get_errnum() << L"] " << ex.get_apiname()
                      << L": " << ex.get_errmsg() << endl;
        return 2;
    }

    return 0;
}
